#include <iostream>
#include "symboltable.h"

SymbolTable::SymbolTable(int initSize) {
	buckets = new list<Symbol>[initSize];
	tableSize = initSize;
}

int SymbolTable::hash(const string &str) {
	int bucketID = str[0];
	for(int i = 1; i < str.size(); i++) {
		bucketID *= 32;
		bucketID += str[i];
		bucketID %= tableSize;
	}
	return bucketID;
}

void SymbolTable::insert(const Symbol &symbol) {
	int bucket = hash(symbol.get_name());

	list<Symbol>::iterator i;
	for(i = buckets[bucket].begin(); i != buckets[bucket].end(); i++) {
		if((*i).get_name() == symbol.get_name()) return;
	}
	buckets[bucket].push_front(symbol);
}

Symbol *SymbolTable::retrieve(const string &name) {
	int bucket = hash(name);

	list<Symbol>::iterator i;
	for(i = buckets[bucket].begin(); i != buckets[bucket].end(); i++) {
		if((*i).get_name() == name) return &*i;
	}
	return NULL;
}

void SymbolTable::erase(const string &name) {
	int bucket = hash(name);
	list<Symbol>::iterator i;
	for(i = buckets[bucket].begin(); i != buckets[bucket].end(); i++) {
		if((*i).get_name() == name) {
			buckets[bucket].erase(i);
			return;
		}
	}
}

void SymbolTable::traverse(VisitFunction func, void *arg) {
	for(int k = 0; k < tableSize; k++) {
		list<Symbol>::iterator i;
		for(i = buckets[k].begin(); i != buckets[k].end(); i++) {
			func(*i, arg);
		}
	}
}

void SymbolTable::clear() {
	for(int i = 0; i < tableSize; i++) {
		buckets[i].clear();
	}
}

void SymbolTable::resize(unsigned int newSize) {
	if(newSize == 0) return;

	list<Symbol> tempHold;

	for(int k = 0; k < tableSize; k++) {
		list<Symbol>::iterator i;
		for(i = buckets[k].begin(); i != buckets[k].end(); i++) {
			tempHold.push_front(*i);
		}
	}
		
	tableSize = newSize;
	buckets = new list<Symbol>[tableSize];

	list<Symbol>::iterator i;
	for(i = tempHold.begin(); i != tempHold.end(); i++) {
		insert(*i);
	}
}

