#include <iostream>
#include "symboltable.h"

using namespace std;

SymbolTable::SymbolTable(int initSize) {
	tableSize = initSize;
	if(tableSize % 2 == 0) {
		tableSize++;
	}		
	items = new Symbol[tableSize];
}

int SymbolTable::hash1(const string &str) {
	int bucketID = str[0];
	for(int i = 1; i < str.size(); i++) {
		bucketID *= 32;
		bucketID += str[i];
		bucketID %= tableSize;
	}
	return bucketID;
}

int SymbolTable::hash2(int n) {
	int u = (tableSize - 1) / 2;
	return u - (n % u);
}

void SymbolTable::insert(const Symbol &symbol) {
	int pos = hash1(symbol.get_name());

	if(items[pos].state == Empty) {
		items[pos] = symbol;
		items[pos].state = Full;
		return;
	} else if(items[pos].get_name() == symbol.get_name() && items[pos].state == Full) {
		return;
	}

	int stepSize = hash2(pos);

	for(int i = pos + stepSize; i != pos; i += stepSize) {
		i %= tableSize;
		if(items[i].state == Empty) {
			items[i] = symbol;
			items[i].state = Full;
			return;
		} else if(items[i].get_name() == symbol.get_name()) {
			return;
		}
	}
}

Symbol *SymbolTable::retrieve(const string &name) {
	int pos = hash1(name);

	if(items[pos].state == Empty) {
		return NULL;
	} else if(items[pos].state == Full && items[pos].get_name() == name) {
		return &(items[pos]);
	}

	int stepSize = hash2(pos);

	for(int i = pos + stepSize; i != pos; i += stepSize) {
		i %= tableSize;
		if(items[i].state == Empty) {
			return NULL;
		} else if(items[i].state == Full && items[i].get_name() == name) {
			return &(items[i]);
		}
	}
	return NULL;
}

void SymbolTable::erase(const string &name) {
	int pos = hash1(name);

	if(items[pos].state == Empty) {
		return;
	} else if(items[pos].get_name() == name && items[pos].state == Full) {
		items[pos].state = Deleted;
		return;
	}

	int stepSize = hash2(pos);

	for(int i = pos; i != pos - 1; i++) {
		i %= tableSize;

		if(items[i].state == Empty) {
			return;
		} else if(items[i].get_name() == name && items[pos].state == Full) {
			items[i].state = Deleted;
			return;
		}
	}
}

void SymbolTable::traverse(VisitFunction func, void *arg) {
	for(int i = 0; i < tableSize; i++) {
		if(items[i].state == Full) func(items[i], arg);
	}
}

void SymbolTable::clear() {
	for(int i = 0; i < tableSize; i++) {
		items[i].state = Empty;
	}	
}

void SymbolTable::resize(int newSize) {
	list<Symbol> tempHold;
	for(int i = 0; i < tableSize; i++) {
		if(items[i].state == Full) tempHold.push_front(items[i]);
	}

	delete[] items;

	tableSize = newSize;
	if(tableSize % 2 == 0) {
		tableSize++;
	}

	items = new Symbol[tableSize];

	list<Symbol>::iterator i = tempHold.begin();
	for(int n = 0; i != tempHold.end(); i++, n++) {
		if(n == tableSize) return;
		insert(*i);
	}
}

