#include <list>
#include <string>

using namespace std;

// just an implementation detail
enum SymbolState { Empty, Full, Deleted };

/* Symbol represents an identifier in program's source code such as a variable
   or function name */
class Symbol {
	string name;
	SymbolState state;
public:
	Symbol() : state(Empty) {};
	Symbol(const string &initName) : name(initName) {};

	// used to access the name of the symbol
	string get_name() const { return name; };
	void set_name(const string &newName) { name = newName; };

	bool operator==(const Symbol &rhs) { return name == rhs.get_name(); };
	friend class SymbolTable;
};

typedef void (*VisitFunction)(Symbol &, void *);

class SymbolTable {
	Symbol *items;

	int tableSize;

	int hash1(const string &);
	int hash2(int);
public:
	// creates an empty table of size 1
	SymbolTable() {};

	// creates an empty table of size n
	SymbolTable(int n);

	// insert a copy of a symbol into the table, will reject duplicates
	void insert(const Symbol &);
	
	// returns a pointer to the matching symbol, NULL if not found
	Symbol *retrieve(const string &);

	// erases the given symbol, does nothing if symbol is not found
	void erase(const string &);

	// executes func(s, arg) for every symbol as s
	void traverse(VisitFunction func, void *arg);

	// clears the entire table
	void clear();

	// sets the table size to newSize
	void resize(int newSize);
};

