#include <iostream>
#include <math.h>
#include "symboltable.h"

SymbolTable::SymbolTable(int initSize) : items(NULL) {
	resize(initSize);
}

int SymbolTable::hash(const string &str) {
	int bucketID = str[0];
	for(int i = 1; i < str.size(); i++) {
		bucketID *= 32;
		bucketID += str[i];
		bucketID %= tableSize;
	}	
	return bucketID;
}

void SymbolTable::insert(const Symbol &symbol) {
	int first = hash(symbol.get_name());

	int firstAgain = first - 1;

	for(int i = 0; (int)pow(i, 2) % tableSize + first != firstAgain; i++) {
		int pos = (int)pow(i, 2) % tableSize + first;

		if(items[pos].state == Empty) {
			items[pos] = symbol;
			items[pos].state = Full;
			return;
		} else if(items[pos].get_name() == symbol.get_name() && items[pos].state == Full) {
			return;
		}
		firstAgain = first;
	}
}

Symbol *SymbolTable::retrieve(const string &name) {
	int first = hash(name);
	int firstAgain = first - 1;

	for(int i = 0; (int)pow(i, 2) % tableSize + first != firstAgain; i++) {
		int pos = (int)pow(i, 2) % tableSize + first;

		if(items[pos].state == Empty) {
			return NULL;
		} else if(items[pos].state == Full && items[pos].get_name() == name) {
			 return &(items[pos]);
		}
		firstAgain = first;
	}
	return NULL;
}

void SymbolTable::erase(const string &name) {
	int first = hash(name);

	for(int i = 0; (int)pow(i, 2) % tableSize != 1; i++) {
		int pos = (int)pow(i, 2) % tableSize + first;

		if(items[pos].state == Empty) {
			return;
		} else if(items[pos].get_name() == name) {
			items[pos].state = Deleted;
			return;
		}
	}
}

void SymbolTable::clear() {
	for(int i = 0; i < tableSize; i++) {
		items[i].state = Empty;
		items[i].set_name("");
	}
}

void SymbolTable::resize(int newSize) {
	if(newSize < 1) return;

	list<Symbol> tempHold;

	bool wasEmpty = !(bool)items;
	if(!wasEmpty) {
		for(int i = 0; i < tableSize; i++) {
			if(items[i].state == Full) {
				tempHold.push_front(items[i]);
			}
		}
		delete[] items;
	}

	if(newSize % 2 == 0) newSize++;
	tableSize = newSize;

	items = new Symbol[newSize];
	for(int i = 0; i < tableSize; i++) {
		items[i].state = Empty;
	}

	if(!wasEmpty) {
		list<Symbol>::iterator i;
		for(i = tempHold.begin(); i != tempHold.end(); i++) {
			insert(*i);
		}
	}
}

void SymbolTable::traverse(VisitFunction func, void *arg) {
	for(int i = 0; i < tableSize; i++) {
		if(items[i].state == Full) func(items[i], arg);
	}
}

